from numbers import Number
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from anndata import AnnData
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from insitupy._core.data import InSituData
from insitupy.dataclasses._utils import _get_cell_layer
from insitupy.utils._checks import check_integer_counts


def plot_qc_metrics(
    data: Union[InSituData, AnnData], # type: ignore #TODO: Expand this to InSituExperiment
    cells_layer: Optional[str] = None,
    show_inset: bool = True,
    inset_fraction: Number = 0.2
    ):
    """
    Plots the QC metrics calculated by sc.pp.calculate_qc_metrics.

    Parameters:
    adata : AnnData
        Annotated data matrix with QC metrics calculated.
    """
    if isinstance(data, AnnData):
        adata = data.copy()
    else:
        # retrieve AnnData from cell layer
        celldata = _get_cell_layer(cells=data.cells, cells_layer=cells_layer)
        adata = celldata.matrix.copy()

    # QC metrics in .obs
    obs_metrics = ['total_counts', 'n_genes_by_counts', 'pct_counts_mt']
    # QC metrics in .var
    var_metrics = ['n_cells_by_counts', 'mean_counts', 'pct_dropout_by_counts', 'total_counts']

    # Check if all metrics exist in .obs
    obs_metrics = [metric for metric in obs_metrics if metric in adata.obs]
    if len(obs_metrics) == 0:
        print("Warning: No .obs metrics found in adata.obs")

    # Check if all metrics exist in .var
    var_metrics = [metric for metric in var_metrics if metric in adata.var]
    if len(var_metrics) == 0:
        print("Warning: No .var metrics found in adata.var")

    ncols = len(obs_metrics + var_metrics) + 1  # Adding one more column for the scatter plot
    nrows = 1
    fig, axes = plt.subplots(nrows, ncols, figsize=(5*ncols, 5*nrows))

    # Add titles to each row
    if len(obs_metrics) > 0:
        axes[0].annotate('.obs Metrics', xy=(0, 0.5), xytext=(-axes[0].yaxis.labelpad - 5, 0),
                            xycoords=axes[0].yaxis.label, textcoords='offset points',
                            size='large', ha='right', va='center', rotation=90, weight='bold')

    if len(var_metrics) > 0:
        axes[3].annotate('.var Metrics', xy=(0, 0.5), xytext=(-axes[3].yaxis.labelpad - 5, 0),
                            xycoords=axes[3].yaxis.label, textcoords='offset points',
                            size='large', ha='right', va='center', rotation=90, weight='bold')

    for i, metric in enumerate(obs_metrics):
        sns.histplot(adata.obs[metric], bins=50, color='skyblue', edgecolor='black', kde=False, ax=axes[i])
        axes[i].set_title(metric)
        axes[i].set_xlabel('Value')
        axes[i].set_ylabel('Frequency')

        if show_inset:
            # Inset histogram
            ax_inset = inset_axes(axes[i], width="40%", height="40%", loc='upper right')
            sns.histplot(adata.obs[metric], bins=100, color='skyblue', edgecolor='black', kde=False, ax=ax_inset)
            ax_inset.set_xlim(adata.obs[metric].min(), adata.obs[metric].max() * inset_fraction)  # Adjust the x-limits as needed
            ax_inset.set_xlabel('')
            ax_inset.set_ylabel('')
            ax_inset.set_yticklabels([])

    for i, metric in enumerate(var_metrics):
        sns.histplot(adata.var[metric], bins=50, color='coral', edgecolor='black', kde=False, ax=axes[3+i])
        axes[3+i].set_title(metric)
        axes[3+i].set_xlabel('Value')
        axes[3+i].set_ylabel('Frequency')

    # Add scatter plot for n_genes_by_counts vs total_counts
    if 'n_genes_by_counts' in obs_metrics and 'total_counts' in obs_metrics:
        scatter_ax = axes[len(obs_metrics)]
        scatter_ax.scatter(adata.obs['n_genes_by_counts'], adata.obs['total_counts'], alpha=0.5, color='skyblue', s=8, edgecolor='black', linewidth=0.2)
        scatter_ax.set_title('n_genes_by_counts vs total_counts')
        scatter_ax.set_xlabel('n_genes_by_counts')
        scatter_ax.set_ylabel('total_counts')

    plt.tight_layout()
    plt.show()

def test_transformations(
    data: InSituData, # type: ignore #TODO: Expand this to InSituExperiment
    cells_layer: Optional[str] = None,
    target_sum: Number = 250,
    layer: Optional[str] = None,
    scale: bool = False,
    assert_integer_counts: bool = True
        ):
    """
    Test normalization and transformation methods by plotting histograms of raw,
    log1p-transformed, and sqrt-transformed counts.

    Args:
        adata (AnnData): Annotated data matrix.
        target_sum (int, optional): Target sum for normalization. Defaults to 1e4.
        layer (str, optional): Layer to use for transformation. Defaults to None.
    """
    # retrieve AnnData from cell layer
    celldata = _get_cell_layer(cells=data.cells, cells_layer=cells_layer)
    adata = celldata.matrix.copy() # copy it to not affect it during the plotting

    # Check if the matrix consists of raw integer counts
    if layer is None:
        if assert_integer_counts:
            check_integer_counts(adata.X)
    else:
        adata.X = adata.layers[layer].copy()
        if assert_integer_counts:
            check_integer_counts(adata.X)

    # get raw counts
    raw_counts = adata.X.copy()

    # Preprocessing according to napari tutorial in squidpy
    sc.pp.normalize_total(adata, target_sum=target_sum)

    # Create a copy of the anndata object for log1p transformation
    adata_log1p = adata.copy()
    sc.pp.log1p(adata_log1p)

    # Create a copy of the anndata object for sqrt transformation
    adata_sqrt = adata.copy()
    try:
        X = adata_sqrt.X.toarray()
    except AttributeError:
        X = adata_sqrt.X
    adata_sqrt.X = np.sqrt(X) + np.sqrt(X + 1)

    if scale:
        sc.pp.scale(adata_log1p)
        sc.pp.scale(adata_sqrt)
        titles = ['Log1p-transformed and scaled counts','Sqrt-transformed and scaled counts']
    else:
        titles = ['Log1p-transformed counts','Sqrt-transformed ounts']

    # Plot histograms
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    axes[0].hist(raw_counts.sum(axis=1), bins=50, color='skyblue', edgecolor='black')
    axes[0].set_title('Raw Counts', fontsize=14)
    axes[0].set_xlabel('Counts', fontsize=12)
    axes[0].set_ylabel('Frequency', fontsize=12)

    axes[1].hist(adata_log1p.X.sum(axis=1), bins=50, color='skyblue', edgecolor='black')
    axes[1].set_title(titles[0], fontsize=14)
    axes[1].set_xlabel('Counts', fontsize=12)
    axes[1].set_ylabel('Frequency', fontsize=12)

    axes[2].hist(adata_sqrt.X.sum(axis=1), bins=50, color='skyblue', edgecolor='black')
    axes[2].set_title(titles[1], fontsize=14)
    axes[2].set_xlabel('Counts', fontsize=12)
    axes[2].set_ylabel('Frequency', fontsize=12)


    plt.tight_layout()
    plt.show()


import os
from math import ceil
from numbers import Number
from pathlib import Path
from typing import List, Optional, Union

import dask.array as da
import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
from anndata import AnnData
from scipy.sparse import csr_matrix
from shapely.geometry import Polygon
from shapely.ops import unary_union

from insitupy.utils._shapely import scale_polygon


def _read_baysor_polygons(
    file: Union[str, os.PathLike, Path]
    ) -> gpd.GeoDataFrame:
    from insitupy._io.files import read_json

    d = read_json(file)

    # prepare output dictionary
    df = {
    "geometry": [],
    "cell": [],
    "type": [],
    "minx": [],
    "miny": [],
    "maxx": [],
    "maxy": []
    }

    for elem in d["geometries"]:
        coords = elem["coordinates"][0]

        # check if there are enough coordinates for a Polygon (some segmented cells are very small in Baysor)
        if len(coords) > 3:
            p = shapely.Polygon(coords)
            df["geometry"].append(p)
            df["type"].append("polygon")

        else:
            p = shapely.LineString(coords)
            df["geometry"].append(p)
            df["type"].append("line")
        df["cell"].append(elem["cell"])

        # extract bounding box
        bounds = p.bounds
        df["minx"].append(bounds[0])
        df["miny"].append(bounds[1])
        df["maxx"].append(bounds[2])
        df["maxy"].append(bounds[3])

    # create geopandas dataframe
    df = gpd.GeoDataFrame(df)

    return df


def read_baysor_transcripts(
    baysor_output: Union[str, os.PathLike, Path]
    ) -> pd.DataFrame:

    # convert to pathlib path
    baysor_output = Path(baysor_output)

    # read transcripts from Baysor results
    print("Parsing transcripts data...", flush=True)

    print("\tRead data", flush=True)
    segcsv_file = baysor_output / "segmentation.csv"
    baysor_transcript_dataframe = pd.read_csv(segcsv_file)

    # reshaping
    transcript_id_col = [elem for elem in ["transcript_id", "molecule_id"] if elem in baysor_transcript_dataframe.columns][0]
    baysor_transcript_dataframe = baysor_transcript_dataframe.set_index(transcript_id_col)
    return baysor_transcript_dataframe


def _read_proseg_polygons(
    file: Union[str, os.PathLike, Path]
    ) -> gpd.GeoDataFrame:
    from insitupy._io.files import read_json

    d = read_json(file)

    # prepare output dictionary
    df = {
    "geometry": [],
    "cell": [],
    "type": [],
    "minx": [],
    "miny": [],
    "maxx": [],
    "maxy": []
    }

    for feature in d['features']:
        geometry = feature['geometry']
        properties = feature["properties"]

        if geometry['type'] == 'MultiPolygon':
            polygons = [Polygon(coords[0]) for coords in geometry['coordinates']]
            merged_geometry = unary_union(polygons).convex_hull
            df["geometry"].append(merged_geometry)
            df["type"].append("polygon")

        elif geometry['type'] == 'Polygon':
            merged_geometry = Polygon(geometry['coordinates'][0])
            df["geometry"].append(merged_geometry)
            df["type"].append("polygon")

        df["cell"].append(properties['cell'])

        # extract bounding box
        bounds = merged_geometry.bounds
        df["minx"].append(bounds[0])
        df["miny"].append(bounds[1])
        df["maxx"].append(bounds[2])
        df["maxy"].append(bounds[3])


    # create geopandas dataframe
    df = gpd.GeoDataFrame(df)

    return df


def _read_proseg_counts(
    path_counts,
    path_cell_metadata,
    exclude_patterns_genes: List[str] = ["NegControl", "Unassigned", "BLANK_", "antisense_"]
    ):
    path_counts = Path(path_counts)
    path_cell_metadata = Path(path_cell_metadata)
    # Read counts data based on file extension
    path_counts_suffix = path_counts.name.split(sep=".", maxsplit=1)[1]
    if path_counts_suffix == "parquet":
        counts = pd.read_parquet(path_counts)
    elif path_counts_suffix == "csv.gz":
        counts = pd.read_csv(path_counts, compression='gzip')
    elif path_counts_suffix == "csv":
        counts = pd.read_csv(path_counts)
    else:
        raise ValueError(f"Unexpected file ending of path_counts: {path_counts_suffix}.")

    # convert counts to float32
    print("Convert counts to float32.", flush=True)
    counts = counts.astype(np.float32)

    # Read metadata based on file extension
    path_metadata_suffix = path_cell_metadata.name.split(sep=".", maxsplit=1)[1]
    if path_metadata_suffix == "parquet":
        meta = pd.read_parquet(path_cell_metadata)
    elif path_metadata_suffix == "csv.gz":
        meta = pd.read_csv(path_cell_metadata, compression='gzip')
    elif path_metadata_suffix == "csv":
        meta = pd.read_csv(path_cell_metadata)
    else:
        raise ValueError(f"Unexpected file ending of path_cell_metadata: {path_metadata_suffix}.")

    # Ensure indices are strings
    counts.index = counts.index.astype(str)
    meta.index = meta.index.astype(str)

    # Filter out unwanted columns
    counts = counts.loc[:, ~counts.columns.str.contains("|".join(exclude_patterns_genes))]

    # Add spatial coordinates
    obsm = {"spatial": np.stack([meta["centroid_x"].to_numpy(), meta["centroid_y"].to_numpy()], axis=1)}

    # Create AnnData object
    adata = AnnData(X=counts, obs=meta, obsm=obsm)

    # make sure the counts are sparse
    adata.X = csr_matrix(adata.X)

    # set the cell column as index
    adata.obs.set_index("cell", inplace=True)
    adata.obs_names = adata.obs_names.astype(str)
    adata.obs_names.name = None # remove the name of the index

    return adata


def _read_proseg(
    path,
    counts_file: Optional[str] = None,
    cell_metadata_file: Optional[str] = None,
    polygons_file: Optional[str] = None,
    pixel_size: Number = 1
    ):
    try:
        from rasterio.features import rasterize
    except ImportError:
        raise ImportError("This function requires the rasterio package, please install with `pip install rasterio`.")

    if counts_file is None:
        path_counts = list(path.glob("expected-counts.*"))[0]
    else:
        path_counts = path / counts_file

    if cell_metadata_file is None:
        path_cell_metadata = list(path.glob("cell-metadata.*"))[0]
    else:
        path_cell_metadata = path / cell_metadata_file

    if polygons_file is None:
        path_polygons = list(path.glob("cell-polygons.*"))[0]
    else:
        path_polygons = path / polygons_file

    # read proseg counts
    adata = _read_proseg_counts(path_counts, path_cell_metadata)

    # Read Proseg polygons
    polygons = _read_proseg_polygons(path_polygons)

    # Scale Baysor polygons
    if pixel_size != 1:
        polygons['geometry'] = polygons['geometry'].apply(lambda x: scale_polygon(x, pixel_size))
        polygons["maxx"] = polygons["maxx"] / pixel_size
        polygons["maxy"] = polygons["maxy"] / pixel_size


    # Calculate bounds for rasterization
    polygon_bounds = polygons.geometry.bounds
    xmax = ceil(polygon_bounds.loc[:, "maxx"].max())
    ymax = ceil(polygon_bounds.loc[:, "maxy"].max())

    # get cell names and generate segmentation mask values
    cell_names = polygons['cell'].values
    seg_mask_value = range(1, len(polygons['cell'])+1)

    # rasterize polygons
    boundaries_mask = rasterize(list(zip(polygons["geometry"], seg_mask_value)), out_shape=(ymax,xmax))
    boundaries_mask = da.from_array(boundaries_mask)

    return adata, boundaries_mask, cell_names, seg_mask_value


"""
Functions in this module were adapted from the decoupler package v1.9.0 (https://github.com/scverse/decoupler):
Badia-i-Mompel P., Vélez Santiago J., Braunger J., Geiss C., Dimitrov D., Müller-Dott S.,
Taus P., Dugourd A., Holland C.H., Ramirez Flores R.O. and Saez-Rodriguez J. 2022.
decoupleR: Ensemble of computational methods to infer biological activities from omics data.
Bioinformatics Advances. https://doi.org/10.1093/bioadv/vbac016

"""


from typing import Literal, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData
from scipy.sparse import csr_matrix, issparse, vstack

from insitupy.dataclasses._utils import _get_cell_layer
from insitupy.experiment.data import InSituExperiment


def psbulk_profile(profile, mode='sum'):
    if mode == 'sum':
        profile = np.sum(profile, axis=0)
    elif mode == 'mean':
        profile = np.mean(profile, axis=0)
    elif mode == 'median':
        profile = np.median(profile, axis=0)
    elif callable(mode):
        profile = np.apply_along_axis(mode, 0, profile)
    else:
        raise ValueError("""mode={0} can be 'sum', 'mean', 'median' or a callable function.""".format(mode))
    return profile

def extract_psbulk_inputs(
    exp: InSituExperiment,
    cells_layer: str,
    layer: str
    ):

    obs_columns_list = []
    for _, data in exp.iterdata():
        celldata = _get_cell_layer(
            cells=data.cells,
            cells_layer=cells_layer
        )
        obs_columns_list.append(set(celldata.matrix.obs.columns))

    common_obs_columns = sorted(set.intersection(*obs_columns_list))

    obs = pd.DataFrame()
    #obs = {}

    # for c in common_obs_columns:
    #     column_dict={}
    #     for metadata, data in exp.iterdata():
    #         celldata = _get_cell_layer(
    #             cells=data.cells,
    #             cells_layer=cells_layer
    #     )
    #         column_dict[metadata["uid"]] = celldata.matrix.obs[c].copy()

    #     column=pd.concat([
    #         s.reset_index(drop=True) for s in column_dict.values()],
    #                      axis=0)
    #     obs[c]=column

    obs_per_data = {}
    for metadata, data in exp.iterdata():
        celldata = _get_cell_layer(
            cells=data.cells,
            cells_layer=cells_layer
        )

        obs_per_data[metadata["uid"]] = celldata.matrix.obs[common_obs_columns]

    obs = pd.concat(obs_per_data).reset_index(level=0, names="batch")

    var_names = []
    for _, data in exp.iterdata():
        celldata = _get_cell_layer(
                    cells=data.cells,
                    cells_layer=cells_layer
                )
        var_names.append(set(celldata.matrix.var_names))

    common_var_names = sorted(set.intersection(*var_names))

    var = pd.DataFrame({"genes": common_var_names})

        # Extract count matrix X
    X_list=[]
    for metadata, data in exp.iterdata():
            celldata = _get_cell_layer(
                        cells=data.cells,
                        cells_layer=cells_layer
                    )
            ad = celldata.matrix
            ad = ad[:, ad.var_names.isin(common_var_names)]
            if layer is not None:
                X_list.append(ad.layers[layer])
            else:
                X_list.append(ad.X)
    X = vstack(X_list)

    # Sort genes
    msk = np.argsort(var["genes"])

    X = X[:, msk]
    var = var.iloc[msk]
    var.set_index('genes', inplace=True)

    if issparse(X) and not isinstance(X, csr_matrix):
        X = csr_matrix(X)

    return X, obs, var

def format_psbulk_inputs(sample_col, groups_col, obs):
    # Use one column if the same
    if sample_col == groups_col:
        groups_col = None

    if groups_col is None:
        # Filter extra columns in obs
        cols = obs.groupby(sample_col, observed=True).nunique(dropna=False).eq(1).all(0)
        cols = np.hstack([sample_col, cols[cols].index])
        obs = obs.loc[:, cols]

        # Get unique samples
        smples = np.unique(obs[sample_col].values)
        groups = None

        # Get number of samples and features
        n_rows = len(smples)
    else:
        # Check if extra grouping is needed
        if type(groups_col) is list:
            obs = obs.copy()
            joined_cols = '_'.join(groups_col)
            obs[joined_cols] = obs[groups_col[0]].str.cat(obs[groups_col[1:]].astype('U'), sep='_')
            groups_col = joined_cols

        # Filter extra columns in obs
        cols = obs.groupby([sample_col, groups_col], observed=True).nunique(dropna=False).eq(1).all(0)
        cols = np.hstack([sample_col, groups_col, cols[cols].index])
        obs = obs.loc[:, cols]

        # Get unique samples and groups
        smples = np.unique(obs[sample_col].values)
        groups = np.unique(obs[groups_col].values)

        # Get number of samples and features
        n_rows = len(smples) * len(groups)

    return obs, groups_col, smples, groups, n_rows
def psbulk_profile(profile, mode='sum'):
    if mode == 'sum':
        profile = np.sum(profile, axis=0)
    elif mode == 'mean':
        profile = np.mean(profile, axis=0)
    elif mode == 'median':
        profile = np.median(profile, axis=0)
    elif callable(mode):
        profile = np.apply_along_axis(mode, 0, profile)
    else:
        raise ValueError("""mode={0} can be 'sum', 'mean', 'median' or a callable function.""".format(mode))
    return profile
def compute_psbulk(n_rows, n_cols, X, sample_col, groups_col, smples, groups, obs,
                   new_obs, min_cells, min_counts, mode, dtype):

    # Init empty variables
    psbulk = np.zeros((n_rows, n_cols))
    props = np.zeros((n_rows, n_cols))
    ncells = np.zeros(n_rows)
    counts = np.zeros(n_rows)

    # Iterate for each group and sample
    i = 0
    if groups_col is None:
        for smp in smples:
            # Write new meta-data
            tmp = obs[obs[sample_col] == smp].drop_duplicates().values
            new_obs.loc[smp, :] = tmp

            # Get cells from specific sample
            profile = X[(obs[sample_col] == smp).values]
            if isinstance(X, csr_matrix):
                profile = profile.toarray()

            # Skip if few cells or not enough counts
            ncell = profile.shape[0]
            count = np.sum(profile)
            ncells[i] = ncell
            counts[i] = count
            if ncell < min_cells or np.abs(count) < min_counts:
                i += 1
                continue

            # Get prop of non zeros
            prop = np.sum(profile != 0, axis=0) / profile.shape[0]

            # Pseudo-bulk
            profile = psbulk_profile(profile, mode=mode)

            # Append
            props[i] = prop
            psbulk[i] = profile
            i += 1
    else:
        for grp in groups:
            for smp in smples:
                # Write new meta-data
                index = smp + '-' + grp
                tmp = obs[(obs[sample_col] == smp) & (obs[groups_col] == grp)].drop_duplicates().values
                if tmp.shape[0] == 0:
                    tmp = np.full(tmp.shape[1], np.nan)
                new_obs.loc[index, :] = tmp

                # Get cells from specific sample and group
                profile = X[((obs[sample_col] == smp) & (obs[groups_col] == grp)).values]
                if isinstance(X, csr_matrix):
                    profile = profile.toarray()

                # Skip if few cells or not enough counts
                ncell = profile.shape[0]
                count = np.sum(profile)
                ncells[i] = ncell
                counts[i] = count
                if ncell < min_cells or np.abs(count) < min_counts:
                    i += 1
                    continue

                # Get prop of non zeros
                prop = np.sum(profile != 0, axis=0) / profile.shape[0]

                # Pseudo-bulk
                profile = psbulk_profile(profile, mode=mode)

                # Append
                props[i] = prop
                psbulk[i] = profile
                i += 1

    return psbulk, ncells, counts, props

def filter_by_prop(adata, min_prop=0.2, min_smpls=2):
    """
    Determine which genes are expressed in a sufficient proportion of cells across samples.

    This function selects genes that are sufficiently expressed across cells in each sample and that this condition is
    met across a minimum number of samples.

    Parameters
    ----------
    adata : AnnData
        AnnData obtained after running ``decoupler.get_pseudobulk``. It requires ``.layer['psbulk_props']``.
    min_prop : float
        Minimum proportion of cells that express a gene in a sample.
    min_smpls : int
        Minimum number of samples with bigger or equal proportion of cells with expression than ``min_prop``.

    Returns
    -------
    genes : ndarray
        List of genes to be kept.
    """

    # Define limits
    min_prop = np.clip(min_prop, 0, 1)
    min_smpls = np.clip(min_smpls, 0, adata.shape[0])

    if isinstance(adata, AnnData):
        layer_keys = adata.layers.keys()
        if 'psbulk_props' in list(layer_keys):
            var_names = adata.var_names.values.astype('U')
            props = adata.layers['psbulk_props']
            if isinstance(props, pd.DataFrame):
                props = props.values

            # Compute n_smpl
            nsmpls = np.sum(props >= min_prop, axis=0)

            # Set features to 0
            msk = nsmpls >= min_smpls
            genes = var_names[msk]
            return genes
    raise ValueError("""adata must be an AnnData object that contains the layer 'psbulk_props'. Please check the function
                     decoupler.get_pseudobulk.""")

def swap_layer(adata, layer_key, X_layer_key='X', inplace=False):
    """
    Swaps an ``adata.X`` for a given layer.

    Swaps an AnnData ``X`` matrix with a given layer. Generates a new object by default.

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    layer_key : str
        ``.layers`` key to place in ``.X``.
    X_layer_key : str, None
        ``.layers`` key where to move and store the original ``.X``. If None, the original ``.X`` is discarded.
    inplace : bool
        If ``False``, return a copy. Otherwise, do operation inplace and return ``None``.

    Returns
    -------
    layer : AnnData, None
        If ``inplace=False``, new AnnData object.
    """

    cdata = None
    if inplace:
        if X_layer_key is not None:
            adata.layers[X_layer_key] = adata.X
        adata.X = adata.layers[layer_key]
    else:
        cdata = adata.copy()
        if X_layer_key is not None:
            cdata.layers[X_layer_key] = cdata.X
        cdata.X = cdata.layers[layer_key]

    return cdata
def check_X(X, mode='sum', skip_checks=False):
    if isinstance(X, csr_matrix):
        is_finite = np.all(np.isfinite(X.data))
    else:
        is_finite = np.all(np.isfinite(X))
    if not is_finite:
        raise ValueError('Data contains non finite values (nan or inf), please set them to 0 or remove them.')
    skip_checks = type(mode) is dict or callable(mode) or skip_checks
    if not skip_checks:
        if isinstance(X, csr_matrix):
            is_positive = np.all(X.data >= 0)
        else:
            is_positive = np.all(X >= 0)
        if not is_positive:
            raise ValueError("""Data contains negative values. Check the parameters use_raw and layers to
            determine if you are selecting the correct matrix. To override this, set skip_checks=True.
            """)
        if mode == 'sum':
            if isinstance(X, csr_matrix):
                is_integer = float(np.sum(X.data)).is_integer()
            else:
                is_integer = float(np.sum(X)).is_integer()
            if not is_integer:
                raise ValueError("""Data contains float (decimal) values. Check the parameters use_raw and layers to
                determine if you are selecting the correct data, which should be positive integer counts when mode='sum'.
                To override this, set skip_checks=True.
                """)
def generate_pseudobulk(
    exp,
    sample_col: str,
    groups_col: str,
    cells_layer: Optional[str] = None,
    counts_layer: str = "counts",
    # use_raw: bool = False,
    mode: Literal["sum", "mean", "median"] = "sum",
    min_cells: int = 10,
    min_counts: int = 1000,
    dtype = np.float32,
    skip_checks: bool = False,
    min_prop=None,
    min_smpls=None,
    remove_empty=True
    ):

    min_cells, min_counts = np.clip(min_cells, 1, None), np.clip(min_counts, 1, None)

    X, obs, var = extract_psbulk_inputs(
        exp=exp,
        cells_layer=cells_layer,
        layer=counts_layer
        )

    check_X(X, mode=mode, skip_checks=skip_checks)

    obs, groups_col, smples, groups, n_rows = format_psbulk_inputs(
        sample_col, groups_col, obs
        )

    n_cols = X.shape[1]
    new_obs = pd.DataFrame(columns=obs.columns)

    if type(mode) is dict:
        psbulks = []
        for l_name in mode:
            func = mode[l_name]
            if not callable(func):
                raise ValueError("""mode requieres a dictionary of layer names and callable functions. The layer {0} does not
                contain one.""".format(l_name))
            else:
                # Compute psbulk
                psbulk, ncells, counts, props = compute_psbulk(n_rows, n_cols, X, sample_col, groups_col, smples, groups, obs,
                                                               new_obs, min_cells, min_counts, func, dtype)
                psbulks.append(psbulk)
        layers = {k: v for k, v in zip(mode.keys(), psbulks)}
        layers['psbulk_props'] = props
    elif type(mode) is str or callable(mode):
        # Compute psbulk
        psbulk, ncells, counts, props = compute_psbulk(n_rows, n_cols, X, sample_col, groups_col, smples, groups, obs,
                                                       new_obs, min_cells, min_counts, mode, dtype)
        layers = {'psbulk_props': props}

    # Add QC metrics
    new_obs['psbulk_cells'] = ncells
    new_obs['psbulk_counts'] = counts

    # Create new AnnData
    psbulk = AnnData(psbulk.astype(dtype), obs=new_obs, var=var, layers=layers)

    # Remove empty samples and features
    if remove_empty:
        msk = psbulk.X == 0
        psbulk = psbulk[~np.all(msk, axis=1), ~np.all(msk, axis=0)].copy()

    # Place first element of mode dict as X
    if type(mode) is dict:
        swap_layer(psbulk, layer_key=list(mode.keys())[0], X_layer_key=None, inplace=True)

    # Filter by genes if not None.
    if min_prop is not None and min_smpls is not None:
        if groups_col is None:
            genes = filter_by_prop(psbulk, min_prop=min_prop, min_smpls=min_smpls)
        else:
            genes = []
            for group in groups:
                g = filter_by_prop(psbulk[psbulk.obs[groups_col] == group], min_prop=min_prop, min_smpls=min_smpls)
                genes.extend(g)
            genes = np.unique(genes)
        psbulk = psbulk[:, genes]

    return psbulk



import os
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt


def save_and_show_figure(
    savepath,
    fig,
    save_only: bool = False,
    show: bool = True,
    dpi_save: int = 300,
    background_color: Optional[str] = None,
    tight: bool = True,
    verbose: bool = False
    ):

    if tight:
        fig.tight_layout()

    if savepath is not None:
        print("Saving figure to file " + str(savepath)) if verbose else None

        # create path if it does not exist
        Path(os.path.dirname(savepath)).mkdir(parents=True, exist_ok=True)

        # save figure
        plt.savefig(savepath, dpi=dpi_save,
                    facecolor=background_color, bbox_inches='tight')
        print("Saved.") if verbose else None
    if save_only:
        plt.close(fig)
    elif show:
        plt.show()
    else:
        return


import math
from typing import Literal, Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import cm
from matplotlib.colors import ListedColormap, rgb2hex
from pandas.api.types import is_bool_dtype, is_numeric_dtype

from insitupy._constants import (DEFAULT_CATEGORICAL_CMAP,
                                 DEFAULT_CONTINUOUS_CMAP)
from insitupy.palettes import CustomPalettes
from insitupy.utils._checks import check_raw


def _extract_color_values(adata, key, raw, layer):
    ## Extract expression data
    # check if plotting raw data
    adata_X, adata_var, adata_var_names = check_raw(
        adata,
        use_raw=raw,
        layer=layer
        )

    # locate gene in matrix and extract values
    if key in adata_var_names:
        idx = adata_var.index.get_loc(key)
        color_values = adata_X[:, idx].copy()
        categorical = False

    elif key in adata.obs.columns:
        color_values = adata.obs[key]
        if is_numeric_dtype(adata.obs[key]):
            categorical = False
        else:
            categorical = True
    else:
        color_values = None
        categorical = None

    return color_values, categorical

def _create_handle_function(mode):
    if mode == "circle":
        def handle_function(color):
            return plt.Line2D([0], [0],
                               marker='o',
                               color='w',
                               markerfacecolor=color,
                               markersize=10,
                               markeredgecolor='black',
                               markeredgewidth=1
                               )
    elif mode == "rectangle":
        def handle_function(color):
            return plt.Line2D([0], [0],
                               marker='_',
                               color='w',
                               markerfacecolor=None,
                               markersize=15,
                               markeredgecolor=color,
                               markeredgewidth=8
                               )
    else:
        raise ValueError("Invalid mode. Choose 'circle' or 'rectangle'.")
    return handle_function

def _add_colorlegend_to_axis(
    color_dict: dict,
    ax: plt.Axes,
    max_per_col: int = 10,
    loc: str = 'center',
    bbox_to_anchor: tuple = (0.5, 0.5),
    title: Optional[str] = None,
    #marker: Optional[str] = 'o',
    mode: Literal["circle", "rectangle"] = "circle",
    remove_axis: bool = True
):
    # create function to create handles
    handle_function = _create_handle_function(mode)

    # Create legend manually
    handles = []
    labels = []

    for label, color in color_dict.items():
        handle = handle_function(color)
        handles.append(handle)
        labels.append(label)

    n_col = max(1, math.ceil(len(labels) / max_per_col))

    legend = ax.legend(
        handles,
        labels,
        loc=loc,
        ncol=n_col,
        frameon=True,
        borderpad=0.4,
        labelspacing=0.5,
        handletextpad=0.8,
        #markerscale=1.5,
        title=title,
        bbox_to_anchor=bbox_to_anchor
        )

    if remove_axis:
        # Hide the axis
        ax.set_axis_off()

def _parse_unique_categories(data):
    # retrieve data
    try:
        unique_categories = data.cat.categories # in case of categorical pandas series
    except AttributeError:
        try:
            unique_categories = data.categories # in case of numpy categories
        except AttributeError:
            data = np.array(data)
            try:
                unique_categories = np.sort(data[~data.isna()].unique())
            except AttributeError:
                try:
                    unique_categories = np.sort(np.unique(data[~np.isnan(data)]))
                except TypeError:
                    #unique_categories = np.sort(np.unique(data))
                    # Convert all elements to strings before sorting
                    unique_categories = np.sort(np.unique(data.astype(str)))

    return unique_categories


def create_cmap_mapping(
    data,
    cmap: Optional[Union[str, ListedColormap]] = None,
    rgba_values: Optional[np.ndarray] = None
    ):
    unique_categories = _parse_unique_categories(data)

    if rgba_values is None:
        if cmap is None:
            pal = CustomPalettes()
            cmap = pal.tab20_mod

        # get colormap if necessary
        if not isinstance(cmap, ListedColormap):
            cmap = plt.get_cmap(cmap)

        len_colormap = cmap.N
        category_to_rgba = {category: cmap(i % len_colormap) for i, category in enumerate(unique_categories)}
    else:
        category_to_rgba = {}
        for v in unique_categories:
            idx = np.argwhere(data == v)[0][0].item()
            category_to_rgba[v] = rgba_values[idx]

    return category_to_rgba


def categorical_data_to_rgba(data,
                             cmap: Union[str, ListedColormap],
                             return_mapping: bool = False,
                             nan_val: tuple = (1,1,1,0),
                             rgba_values: Optional[np.ndarray] = None
                             ):

    # len_colormap = cmap.N
    # category_to_rgba = {category: cmap(i % len_colormap) for i, category in enumerate(unique_categories)}

    if not isinstance(cmap, dict):
        category_to_rgba = create_cmap_mapping(data, cmap, rgba_values)
    else:
        category_to_rgba = cmap

    if nan_val is not None:
        # add key for nan
        category_to_rgba[str(np.nan)] = nan_val

    res = np.array([category_to_rgba[str(category)] for category in data])

    if return_mapping:
        return res, category_to_rgba
    else:
        return res


def _determine_climits(
    color_values,
    upper_climit_pct,
    lower_climit = None,
    force_above_zero = False
    ) -> list:

    if lower_climit is None:
        lower_climit = color_values.min()

    if force_above_zero:
        color_values_for_calc = color_values[color_values > 0]
    else:
        color_values_for_calc = color_values
    try:
        upper_climit = np.percentile(color_values_for_calc, upper_climit_pct)
    except IndexError:
        # if there were no color values above zero, a IndexError appears
        upper_climit = 0

    climits = [lower_climit, upper_climit]

    return climits


def continuous_data_to_rgba(
    data,
    cmap: Union[str, ListedColormap],
    upper_climit_pct: int = 99,
    lower_climit: Optional[int] = None,
    clip = False,
    nan_val: tuple = (1,1,1,0),
    return_mapping: bool = False
    ):
    if np.any(pd.isna(data)):
        contains_nans = True
        # Convert the numpy array to a pandas Series
        value_series = pd.Series(data)

        # check where there are nas
        isna_mask = value_series.isna()
        notna_mask = ~isna_mask

        # get only values without NaNs
        notna_values = value_series[notna_mask].values
    else:
        notna_values = data
        contains_nans = False

    # get colormap if necessary
    if not isinstance(cmap, ListedColormap):
        cmap = plt.get_cmap(cmap)

    if lower_climit is None:
        lower_climit = np.min(notna_values)

    climits = _determine_climits(color_values=notna_values, upper_climit_pct=upper_climit_pct, lower_climit=lower_climit)

    norm = mpl.colors.Normalize(vmin=climits[0], vmax=climits[1], clip=clip)
    scalarMap = cm.ScalarMappable(norm=norm, cmap=cmap)
    rgba_data = scalarMap.to_rgba(notna_values)

    if contains_nans:
        result_values = pd.Series(index=value_series.index, dtype="object")
        tuple_list = [tuple(elem) for elem in rgba_data]
        result_values.loc[notna_mask] = tuple_list
        result_values.loc[isna_mask] = [nan_val] * isna_mask.sum()
        result_rgba = np.array(result_values.tolist())
    else:
        result_rgba = rgba_data

    if return_mapping:
        return result_rgba, scalarMap
    else:
        return result_rgba


def _data_to_rgba(
    data: np.ndarray,
    continuous_cmap: Union[str, ListedColormap] = DEFAULT_CONTINUOUS_CMAP,
    categorical_cmap: Union[str, ListedColormap] = None,
    upper_climit_pct: int = 99,
    #return_all: bool = False,
    nan_val: tuple = (1,1,1,0),
    rgba_values: Optional[np.ndarray] = None
    ):
    if not is_numeric_dtype(data) or is_bool_dtype(data):
        if is_bool_dtype(data):
            # in case of boolean data it is important to convert it to object type
            data = data.astype('object')
            # also pandas NAs have to be substituted with numpy NaNs
            data = np.array([elem if pd.notna(elem) else np.nan for elem in data], dtype='object')
        if categorical_cmap is None:
            # pal = CustomPalettes()
            # categorical_cmap = pal.tab20_mod
            categorical_cmap = DEFAULT_CATEGORICAL_CMAP
        rgba_list, mapping = categorical_data_to_rgba(data=data, cmap=categorical_cmap,
                                        return_mapping=True,
                                        nan_val=nan_val,
                                        rgba_values=rgba_values)
        cmap = categorical_cmap
    else:
        rgba_list, mapping = continuous_data_to_rgba(data=data, cmap=continuous_cmap,
                                       upper_climit_pct=upper_climit_pct,
                                       return_mapping=True)
        cmap = continuous_cmap

    return rgba_list, mapping, cmap


def _rgb2hex_robust(rgb, scale_to_one: bool, max_value: int = 255):
    """
    Convert RGB values to hex format, ensuring that the values are within the range [0, 1].
    """
    if scale_to_one:
        rgb = [elem / max_value for elem in rgb]
    # Ensure that the values are within the range [0, 1]
    return rgb2hex(rgb)

# def get_crange(adata, key, use_raw,
#     layer=None,
#     ctype='minmax', cmin_at_zero=True
#     ):
#     obs = adata.obs
#     adata_X, adata_var, adata_var_names = check_raw(adata, use_raw=use_raw, layer=layer)

#     if key in adata_var_names:
#         #c = adata_X[[elem in groups for elem in adata.obs[groupby]], adata_var_names == key]
#         c = adata_X[:, adata_var_names == key]
#         if cmin_at_zero:
#             cmin = 0
#         else:
#             cmin = c.min()

#         if ctype == 'percentile':
#             cmax = np.percentile(c, 95)
#         else:
#             cmax = c.max()
#         crange = [cmin, cmax]
#     elif key in obs.columns:
#         #if obs[key].dtype.name.startswith('float') or obs[key].dtype.name.startswith('int'):
#         if is_numeric_dtype(obs[key]):
#             #c = obs[key][[elem in groups for elem in obs[groupby]]]
#             c = obs[key]
#             if cmin_at_zero:
#                 cmin = 0
#             else:
#                 cmin = c.min()
#             cmax = np.percentile(c, 95)
#             crange = [cmin, cmax]
#         else:
#             return None
#     else:
#         print(f"Key {key} not in var_names of adata object. Use raw?")
#         return

#     return crange